%{
===========================================================================
Code provided by Yichuan (Charlie) Tang
Permission is granted for anyone to copy, use, modify, or distribute this
program and accompanying programs and documents for any purpose, provided
this copyright notice is retained and prominently displayed, along with
a note saying that the original programs are available from our
web page.
The programs and documents are distributed without any warranty, express or
implied.  As the programs were written for research purposes only, they 
have not been tested to the degree that would be advisable in any important
application.  All use of these programs is entirely at the user's own risk.
===========================================================================

CT 3/2012 - modified to be able to handle more than 2 sets of factors
this function returns the log-likelihood of data X under MTA
X - D by N
MoTA - mixture of Tensor Analyzer struct
nSamples - number of Monte Carlo Samples of z1

output:
LogPxGc - C by N - log probability - log p(x|c)
%}
function [ LogPxGc ] = ll_mta3(X, MoTA, nSamples )

[D N] = size(X); %number of samples
C = length(MoTA.mix);

LogPxGcSamples = zeros(C, N, nSamples);

%choose the ind with the largest factors to integrate out
setinds=cell(1,2);

[val setinds{2}] = max(MoTA.d);
setinds{1} = setdiff(1:length(MoTA.d), setinds{2});

% inds into the W required.
this_z_inds = false(1, sum(MoTA.d));
rest_z_inds = false(1, sum(MoTA.d));

ind = 1;
for b = 1:length(MoTA.d)
    if b == setinds{2}
        this_z_inds(ind:ind+MoTA.d(b)-1) = true;
        rest_z_inds(ind:ind+MoTA.d(b)-1) = false;
    else
        this_z_inds(ind:ind+MoTA.d(b)-1) = false;
        rest_z_inds(ind:ind+MoTA.d(b)-1) = true;
    end
    ind = ind + MoTA.d(b);
end

%generate random samples for all latent factors (but won't use d1)
%generate some random samples
z_samples = cell(length(MoTA.d), C);
for dd = 1:size(z_samples,1)
    for cc = 1:size(z_samples,2)
        lat_dim = MoTA.d(dd);
        z_samples{dd,cc} = randn(lat_dim, nSamples );
    end
end

for c = 1:C
    
    Wconcate  = MoTA.W(:,:,c); %concatenated W 
    %split W:
    W_this_z = Wconcate(:, this_z_inds);    %2-biases
    W_rest_z = Wconcate(:, rest_z_inds);    %to be multiplied with z_given
    
    y = cat(1, z_samples{setinds{1},c});
    
    psi   = MoTA.Psi(:,c);    
    psiI  = psi.^-1;
    
    z_c = cell(length(MoTA.d),1);
    [z_c{:}] = z_samples{:,c};
    W_batch = bsxfun(@plus, W_this_z, TensorMatrixProd_zs( MoTA.T(:,:,c), MoTA.d, z_c, setinds{1}, setinds{2} ));
    M = bsxfun(@plus, MoTA.M(:,c), W_rest_z*y);
        
    psiIw = bsxfun(@times, W_batch, psiI);
    
    thisd = MoTA.d(setinds{2});
    
    if 1
    
        parfor n = 1:nSamples       
            W = W_batch(:,:,n);

            lat_precision = eye(thisd) + W'* psiIw(:,:,n); % V_c        
            diff = bsxfun(@minus, X, M(:,n));
            psiI_w_xc  = psiIw(:,:,n)'*diff;

            L=chol(lat_precision);
            log_det_lat_cov = -2*sum(log(diag(L)));

            log_det = sum(log(psi)) - log_det_lat_cov;

            energy = -.5*psiI'*(diff.*diff) ;
            energy = energy+.5*sum( psiI_w_xc .* ( lat_precision \ psiI_w_xc),1);
            LogPxGcSamples(c,:,n) = energy -.5*log_det -(D/2)*log(2*pi);

        end %nSamples
    
    else
        Diff = repmat(X, [1 1 nSamples])-repmat( reshape(M, [D, 1, nSamples]), [1, N, 1]);  %64 500, 1000
        Ene = -0.5*psiI'*reshape(Diff.*Diff, [D, N*nSamples]);
        Ene = reshape(Ene,  [1, N, nSamples]);
        
        for n = 1:nSamples
            
            W = W_batch(:,:,n);
            
            lat_precision = eye(thisd) + W'* psiIw(:,:,n); % V_c           
            psiI_w_xc  = psiIw(:,:,n)'*Diff(:,:,n);
            
            L=chol(lat_precision);
            log_det_lat_cov = -2*sum(log(diag(L)));
            
            log_det = sum(log(psi)) - log_det_lat_cov;
            
            energy = Ene(1,:,n)+.5*sum( psiI_w_xc .* ( lat_precision \ psiI_w_xc),1);
            LogPxGcSamples(c,:,n) = energy -.5*log_det -(D/2)*log(2*pi);
            
        end %nSamples        
    end
    
end %C

LogPxGc = logsumexp(LogPxGcSamples+log(1/nSamples), 3);
assert(all(isfinite(LogPxGc(:))));
