%{

===========================================================================
Code provided by Yichuan (Charlie) Tang
Permission is granted for anyone to copy, use, modify, or distribute this
program and accompanying programs and documents for any purpose, provided
this copyright notice is retained and prominently displayed, along with
a note saying that the original programs are available from our
web page.
The programs and documents are distributed without any warranty, express or
implied.  As the programs were written for research purposes only, they 
have not been tested to the degree that would be advisable in any important
application.  All use of these programs is entirely at the user's own risk.
===========================================================================

   CT 3/2012 - adapted to be able to handel more than 2 sets of latent
    factors
 % E-step - this performs MCMC E-step of the inference for MoTA
    X, X2 - D by N data matrix
    nGibbsSample - number of gibbs samples to run

    outputs:
    Q - posterior p(c|x) C by N matrix
    z1 is - [d1 by N by C]
    z2 is - [d2 by N by C]
    LogL  - 1 by N matrix of log likelihoods log p(x)

    %labels - cell{d}
    zs{c,d} is a 2d cell of samples, each cell has dim by N samples.
%}
function [Q, LogL, zs] = mta3_E_step( X, MoTA, nGibbsSamples, ...
                                nLLSamples, labels)

[D N] = size(X);
C = length(MoTA.mix);
zs = cell(C, length(MoTA.d));
for cc = 1:C
    for dd =1:size(zs,2)
        zs{cc,dd} = zeros(MoTA.d(dd), N);
    end
end

for i = 1:nGibbsSamples    
   
    fprintf( ' %d.', i);
    for dd = 1:length(MoTA.d)
         
        %fprintf( '\nmta3Estep %d/%d dd:%d  ', i, nGibbsSamples, dd);
        %pause(0.1);
        
        setinds = [];
        setinds{1} = [1:dd-1 dd+1:length(MoTA.d)];
        setinds{2} = dd;   
        
        if exist('labels', 'var') && ~isempty(labels) && ~isempty(labels{dd})
            
            [ LogL, zs ] = postinf_mta_given_zs_labels( X, MoTA, zs, setinds, labels{dd});
            
            if min(labels{dd}) < 1 %do semi-supervised learning
                [ LogL, zs_supplement ] = postinf_mta_given_zs( X, MoTA, zs, setinds );
                
                nolabels_inds = labels{dd} < 1;
                for c=1:C
                    zs{c,dd}(:,nolabels_inds) =  zs_supplement{c,dd}(:,nolabels_inds);
                end
            end
        else
            [ LogL, zs ] = postinf_mta_given_zs( X, MoTA, zs, setinds );
        end        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%        
    end
end

fprintf(' Estep(ll_mta3),');
LogPxGc  = ll_mta3(X, MoTA, nLLSamples );


assert(all(isfinite(LogPxGc(:))));

LogPxc = bsxfun(@plus, LogPxGc, log(MoTA.mix+realmin));

Q    = bsxfun(@minus, LogPxc, max(LogPxc,[],1) );
Q    = exp(Q)+realmin;
Q    = bsxfun(@rdivide, Q , sum(Q,1) );
LogL = sum(Q.*(-log(Q+realmin) + LogPxc),1); %lower bound (which is tight, since Q is exactly the posterior)

assert(all(isfinite(LogL)));

