%{

===========================================================================
Code provided by Yichuan (Charlie) Tang
Permission is granted for anyone to copy, use, modify, or distribute this
program and accompanying programs and documents for any purpose, provided
this copyright notice is retained and prominently displayed, along with
a note saying that the original programs are available from our
web page.
The programs and documents are distributed without any warranty, express or
implied.  As the programs were written for research purposes only, they 
have not been tested to the degree that would be advisable in any important
application.  All use of these programs is entirely at the user's own risk.
===========================================================================

 function [ LogL, zs ] = postinf_mta_given_zs( X, MoTA, z2, X2 )
  
 purpose is to perform posterior inference in the MTA
 general purpose function to sample a set z, given all other sets of z's


 X (D x N)     - n d-dimensional data
 
 zs{c,d} is a 2d cell of samples, each cell has dim by N samples. 
 e.g. zs{C,2} is [d2 by N] matrix of projections  (a random sample)
 setinds - cell 1. row vector of conditioned latent factor sets. (rest)
                2. latent factor sets to sample (this)
 e.g. setinds{1} = [2 3], setinds{2} = [1]

 mta is a struct with the fields:
     M (D x C)     - k d-dimensional means
     W (D x (d1+d2+d3) x C) - 2-biases
     Psi (D,C)     - diagonal noise levels
     T - tensor (matricized) D x (d1*d2) x C
     mix - C x 1 mixing proportions
     .d - dimensionality = [d1=2 d2=5 d3=10]
 
% LogL is C by N matrix  log p(x|zs, c)
% z is d(?) by N by C matrix of projections  (a random sample)
%}

function [ LogL, zs] = postinf_mta_given_zs_labels( X, MoTA, zs, setinds, labels, bNoise)

if ~exist('bNoise', 'var')
    bNoise = true;
end

N = size(X,2); %number of samples
[D, ~, C] = size(MoTA.T);

%do some error checking
assert( all( size(labels) == [1 N]) );

assert(size(X,1) == D);
assert(length(setinds) ==2);
assert(length(setinds{2}) == 1);
assert( length([setinds{1} setinds{2}]) == length(MoTA.d));
assert( sum([setinds{1} setinds{2}]) == (length(MoTA.d)*(length(MoTA.d)+1))/2);

assert(size(zs,2) ==length(MoTA.d));
assert(size(zs,1) == C);

LogL = zeros(C,N);

% inds into the W required.
this_z_inds = false(1, sum(MoTA.d));
rest_z_inds = false(1, sum(MoTA.d));

ind = 1;
for b = 1:length(MoTA.d)
    if b == setinds{2}
        this_z_inds(ind:ind+MoTA.d(b)-1) = true;      
        rest_z_inds(ind:ind+MoTA.d(b)-1) = false;
    else
        this_z_inds(ind:ind+MoTA.d(b)-1) = false;
        rest_z_inds(ind:ind+MoTA.d(b)-1) = true;
    end
    ind = ind + MoTA.d(b);
end

for c = 1:C
    
    Wconcate  = MoTA.W(:,:,c); %concatenated W 
    %split W:
    W_this_z = Wconcate(:, this_z_inds);    %2-biases
    W_rest_z = Wconcate(:, rest_z_inds);    %to be multiplied with z_given
    
    y = cat(1, zs{c,setinds{1}});
    
    psi   = MoTA.Psi(:,c);    
    psiI  = psi.^-1;
        
    z_c = cell(length(MoTA.d),1);
    [z_c{:}] = zs{c,:};
    W_batch = bsxfun(@plus, W_this_z, TensorMatrixProd_zs( MoTA.T(:,:,c), MoTA.d, z_c, setinds{1}, setinds{2} ));
    M = bsxfun(@plus, MoTA.M(:,c), W_rest_z*y);
    %W, m are the MFA parameters conditioned on z2
    assert( size(W_batch,3) == N && size(M,2) == N);
    
    psiIw = bsxfun(@times, W_batch, psiI);
    
    for ss =1:max(labels(:))
        
        inds = find(labels==ss);
        if ~isempty(inds)        

            lat_precision = eye(MoTA.d(setinds{2}));
            psiI_w_xc = zeros(MoTA.d(setinds{2}),1);
            
            for bb = 1:length(inds)
                W = W_batch(:,:, inds(bb) );
                
                lat_precision = lat_precision + W'*psiIw(:,:,inds(bb)); % V_c
                
                diff =  X(:,inds(bb))- M(:,inds(bb));
                psiI_w_xc  = psiI_w_xc + psiIw(:,:,inds(bb))'*diff;                                
            end
            
            z_mean = lat_precision \ psiI_w_xc;
          
            %log_det_lat_cov = -2*sum(log(diag(R)));
            if bNoise
                  
                R=chol(lat_precision);
                zs{c,setinds{2}}(:,inds) = repmat(z_mean + R\randn(size(z_mean)), [1 length(inds)]);
                %needs inv(R)*randn b/c R'*R == lat_precision, not lat_cov
            else
                zs{c,setinds{2}}(:,inds) = repmat(z_mean, [1 length(inds)]);
            end

            %below not really needed
            %log_det = sum(log(psi)) - log_det_lat_cov;        
            %energy = -.5*psiI'*(diff.*diff) +.5 * sum( psiI_w_xc .* ( lat_precision \ psiI_w_xc),1);
            %LogL(c,n) = energy -.5*log_det -(D/2)*log(2*pi);        
        end    
    end
end

assert(all(isfinite(LogL(:))));
